@@ -1,11 +1,20 @@
-#include "pgroonga.h"
+#include <groonga.h>
+
+#define PGRN_TAG "pgroonga: crash-safer"
+
+static grn_ctx *ctx = NULL;
+
 #include "pgrn-compatible.h"
 #include "pgrn-portable.h"
 
 #include "pgrn-crash-safer-statuses.h"
 #include "pgrn-database.h"
 #include "pgrn-file.h"
+#include "pgrn-log-level.h"
 #include "pgrn-value.h"
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+#	include "pgrn-wal-custom.h"
+#endif
 
 #include <access/heapam.h>
 #include <access/tableam.h>
@@ -19,17 +28,15 @@
 #include <storage/ipc.h>
 #include <storage/latch.h>
 #include <storage/procsignal.h>
-#include <utils/snapmgr.h>
 #include <utils/guc.h>
 #include <utils/snapmgr.h>
 #include <utils/timestamp.h>
 
 #include <groonga.h>
 
-
 /* #define PGROONGA_CRASH_SAFER_DEBUG */
 #ifdef PGROONGA_CRASH_SAFER_DEBUG
-#	define P(...) ereport(LOG, (errmsg(TAG __VA_ARGS__)))
+#	define P(...) ereport(LOG, (errmsg(PGRN_TAG __VA_ARGS__)))
 #else
 #	define P(...)
 #endif
@@ -37,16 +44,14 @@
 PG_MODULE_MAGIC;
 
 extern PGDLLEXPORT void _PG_init(void);
-extern PGDLLEXPORT void
-pgroonga_crash_safer_reset_position_one(Datum datum) pg_attribute_noreturn();
-extern PGDLLEXPORT void
-pgroonga_crash_safer_reindex_one(Datum datum) pg_attribute_noreturn();
-extern PGDLLEXPORT void
-pgroonga_crash_safer_flush_one(Datum datum) pg_attribute_noreturn();
-extern PGDLLEXPORT void
-pgroonga_crash_safer_main(Datum datum) pg_attribute_noreturn();
-
-#define TAG "pgroonga: crash-safer"
+extern PGDLLEXPORT void pgroonga_crash_safer_reset_position_one(Datum datum)
+	pg_attribute_noreturn();
+extern PGDLLEXPORT void pgroonga_crash_safer_reindex_one(Datum datum)
+	pg_attribute_noreturn();
+extern PGDLLEXPORT void pgroonga_crash_safer_flush_one(Datum datum)
+	pg_attribute_noreturn();
+extern PGDLLEXPORT void pgroonga_crash_safer_main(Datum datum)
+	pg_attribute_noreturn();
 
 static volatile sig_atomic_t PGroongaCrashSaferGotSIGTERM = false;
 static volatile sig_atomic_t PGroongaCrashSaferGotSIGHUP = false;
@@ -58,7 +63,6 @@ static int PGroongaCrashSaferMaxRecoveryThreads = 0;
 PGRN_DEFINE_LOG_LEVEL_ENTRIES(PGroongaCrashSaferLogLevelEntries);
 static const char *PGroongaCrashSaferLibraryName = "pgroonga_crash_safer";
 
-
 #if PG_VERSION_NUM < 140000
 /* Borrowed from src/backend/utils/adt/timestamp.c in PostgreSQL.
  *
@@ -77,11 +81,10 @@ PGrnTimestampDifferenceMilliseconds(TimestampTz start_time,
 		return (long) ((diff + 999) / 1000);
 }
 #else
-#	define PGrnTimestampDifferenceMilliseconds(start_time, stop_time)	\
-	TimestampDifferenceMilliseconds((start_time), (stop_time))
+#	define PGrnTimestampDifferenceMilliseconds(start_time, stop_time)         \
+		TimestampDifferenceMilliseconds((start_time), (stop_time))
 #endif
 
-
 static uint32_t
 pgroonga_crash_safer_get_thread_limit(void *data)
 {
@@ -91,7 +94,7 @@ pgroonga_crash_safer_get_thread_limit(void *data)
 static void
 pgroonga_crash_safer_sigterm(SIGNAL_ARGS)
 {
-	int	save_errno = errno;
+	int save_errno = errno;
 
 	PGroongaCrashSaferGotSIGTERM = true;
 	SetLatch(MyLatch);
@@ -102,7 +105,7 @@ pgroonga_crash_safer_sigterm(SIGNAL_ARGS)
 static void
 pgroonga_crash_safer_sighup(SIGNAL_ARGS)
 {
-	int	save_errno = errno;
+	int save_errno = errno;
 
 	PGroongaCrashSaferGotSIGHUP = true;
 	SetLatch(MyLatch);
@@ -125,10 +128,8 @@ pgroonga_crash_safer_prepare_one_on_exit(int code, Datum databaseInfoDatum)
 	Oid databaseOid;
 	Oid tableSpaceOid;
 	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
-	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
-											  databaseOid,
-											  tableSpaceOid,
-											  InvalidPid);
+	pgrn_crash_safer_statuses_set_prepare_pid(
+		NULL, databaseOid, tableSpaceOid, InvalidPid);
 }
 
 void
@@ -145,12 +146,10 @@ pgroonga_crash_safer_reset_position_one(Datum databaseInfoDatum)
 	StartTransactionCommand();
 	SPI_connect();
 	PushActiveSnapshot(GetTransactionSnapshot());
-	pgstat_report_activity(STATE_RUNNING, TAG ": resetting position");
+	pgstat_report_activity(STATE_RUNNING, PGRN_TAG ": resetting position");
 
-	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
-											  databaseOid,
-											  tableSpaceOid,
-											  MyProcPid);
+	pgrn_crash_safer_statuses_set_prepare_pid(
+		NULL, databaseOid, tableSpaceOid, MyProcPid);
 	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
 					  databaseInfoDatum);
 
@@ -158,18 +157,19 @@ pgroonga_crash_safer_reset_position_one(Datum databaseInfoDatum)
 		int result;
 
 		SetCurrentStatementStartTimestamp();
-		result = SPI_execute("SELECT proname "
-							 "  FROM pg_catalog.pg_proc "
-							 "  WHERE "
-							 "    proname = 'pgroonga_wal_set_applied_position'",
-							 true,
-							 0);
+		result =
+			SPI_execute("SELECT proname "
+						"  FROM pg_catalog.pg_proc "
+						"  WHERE "
+						"    proname = 'pgroonga_wal_set_applied_position'",
+						true,
+						0);
 		if (result != SPI_OK_SELECT)
 		{
 			ereport(FATAL,
-					(errmsg(TAG ": failed to detect "
-							"pgroonga_wal_set_applied_position(): "
-							"%u/%u: %d",
+					(errmsg(PGRN_TAG ": failed to detect "
+									 "pgroonga_wal_set_applied_position(): "
+									 "%u/%u: %d",
 							databaseOid,
 							tableSpaceOid,
 							result)));
@@ -178,18 +178,18 @@ pgroonga_crash_safer_reset_position_one(Datum databaseInfoDatum)
 		if (SPI_processed > 0)
 		{
 			SetCurrentStatementStartTimestamp();
-			result = SPI_execute("SELECT pgroonga_wal_set_applied_position()",
-								 false,
-								 0);
+			result = SPI_execute(
+				"SELECT pgroonga_wal_set_applied_position()", false, 0);
 			if (result != SPI_OK_SELECT)
 			{
-				ereport(FATAL,
-						(errmsg(TAG ": failed to reset WAL applied positions "
-								"of all PGroonga indexes: "
-								"%u/%u: %d",
-								databaseOid,
-								tableSpaceOid,
-								result)));
+				ereport(
+					FATAL,
+					(errmsg(PGRN_TAG ": failed to reset WAL applied positions "
+									 "of all PGroonga indexes: "
+									 "%u/%u: %d",
+							databaseOid,
+							tableSpaceOid,
+							result)));
 			}
 		}
 	}
@@ -217,12 +217,10 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 	StartTransactionCommand();
 	SPI_connect();
 	PushActiveSnapshot(GetTransactionSnapshot());
-	pgstat_report_activity(STATE_RUNNING, TAG ": reindexing");
+	pgstat_report_activity(STATE_RUNNING, PGRN_TAG ": reindexing");
 
-	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
-											  databaseOid,
-											  tableSpaceOid,
-											  MyProcPid);
+	pgrn_crash_safer_statuses_set_prepare_pid(
+		NULL, databaseOid, tableSpaceOid, MyProcPid);
 	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
 					  databaseInfoDatum);
 
@@ -234,32 +232,33 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 		char **indexNames;
 
 		SetCurrentStatementStartTimestamp();
-		result = SPI_execute("SELECT (namespace.nspname || "
-							 "        '.' || "
-							 "        class.relname) AS index_name "
-							 "  FROM pg_catalog.pg_class AS class "
-							 "  JOIN pg_catalog.pg_namespace AS namespace "
-							 "    ON class.relnamespace = namespace.oid "
-							 " WHERE class.relam = ("
-							 "   SELECT oid "
-							 "     FROM pg_catalog.pg_am "
-							 "    WHERE amname = 'pgroonga'"
-							 " )"
-							 "ORDER BY "
-							 "  CASE "
-							 "  WHEN array_to_string(class.reloptions, ' ', ' ') "
-							 "       LIKE '%${%}%' "
-							 "    THEN 1 "
-							 "  ELSE 0 "
-							 "  END, "
-							 "  class.relname",
-							 true,
-							 0);
+		result =
+			SPI_execute("SELECT (namespace.nspname || "
+						"        '.' || "
+						"        class.relname) AS index_name "
+						"  FROM pg_catalog.pg_class AS class "
+						"  JOIN pg_catalog.pg_namespace AS namespace "
+						"    ON class.relnamespace = namespace.oid "
+						" WHERE class.relam = ("
+						"   SELECT oid "
+						"     FROM pg_catalog.pg_am "
+						"    WHERE amname = 'pgroonga'"
+						" )"
+						"ORDER BY "
+						"  CASE "
+						"  WHEN array_to_string(class.reloptions, ' ', ' ') "
+						"       LIKE '%${%}%' "
+						"    THEN 1 "
+						"  ELSE 0 "
+						"  END, "
+						"  class.relname",
+						true,
+						0);
 		if (result != SPI_OK_SELECT)
 		{
 			ereport(FATAL,
-					(errmsg(TAG ": failed to detect PGroonga indexes: "
-							"%u/%u: %d",
+					(errmsg(PGRN_TAG ": failed to detect PGroonga indexes: "
+									 "%u/%u: %d",
 							databaseOid,
 							tableSpaceOid,
 							result)));
@@ -273,10 +272,8 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 			bool isNull;
 			Datum indexName;
 
-			indexName = SPI_getbinval(SPI_tuptable->vals[i],
-									  SPI_tuptable->tupdesc,
-									  1,
-									  &isNull);
+			indexName = SPI_getbinval(
+				SPI_tuptable->vals[i], SPI_tuptable->tupdesc, 1, &isNull);
 			if (isNull)
 			{
 				indexNames[i] = NULL;
@@ -293,6 +290,20 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 			if (!indexNames[i])
 				continue;
 
+			resetStringInfo(&buffer);
+			appendStringInfo(&buffer,
+							 "SELECT pgroonga_command('log_put', "
+							 "ARRAY["
+							 "'level', 'notice', "
+							 "'message', '%s: reindexing: %s: %u/%u'"
+							 "])",
+							 PGRN_TAG,
+							 indexNames[i],
+							 databaseOid,
+							 tableSpaceOid);
+			SetCurrentStatementStartTimestamp();
+			SPI_execute(buffer.data, false, 0);
+
 			resetStringInfo(&buffer);
 			appendStringInfo(&buffer, "REINDEX INDEX %s", indexNames[i]);
 			SetCurrentStatementStartTimestamp();
@@ -300,13 +311,28 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 			if (result != SPI_OK_UTILITY)
 			{
 				ereport(FATAL,
-						(errmsg(TAG ": failed to reindex PGroonga index: "
-								"%u/%u: <%s>: %d",
+						(errmsg(PGRN_TAG ": failed to reindex PGroonga index: "
+										 "%u/%u: <%s>: %d",
 								databaseOid,
 								tableSpaceOid,
 								indexNames[i],
 								result)));
 			}
+
+			resetStringInfo(&buffer);
+			appendStringInfo(&buffer,
+							 "SELECT pgroonga_command('log_put', "
+							 "ARRAY["
+							 "'level', 'notice', "
+							 "'message', '%s: reindexed: %s: %u/%u'"
+							 "])",
+							 PGRN_TAG,
+							 indexNames[i],
+							 databaseOid,
+							 tableSpaceOid);
+			SetCurrentStatementStartTimestamp();
+			SPI_execute(buffer.data, false, 0);
+
 			pfree(indexNames[i]);
 			indexNames[i] = NULL;
 		}
@@ -332,11 +358,8 @@ pgroonga_crash_safer_flush_one_remove_pid_on_exit(int code,
 	bool found;
 	pgrn_crash_safer_statuses_entry *entry;
 	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
-	entry = pgrn_crash_safer_statuses_search(NULL,
-											 databaseOid,
-											 tableSpaceOid,
-											 HASH_FIND,
-											 &found);
+	entry = pgrn_crash_safer_statuses_search(
+		NULL, databaseOid, tableSpaceOid, HASH_FIND, &found);
 	if (!found)
 		return;
 	entry->pid = InvalidPid;
@@ -380,18 +403,14 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
 
 	databasePath = GetDatabasePath(databaseOid, tableSpaceOid);
-	join_path_components(pgrnDatabasePath,
-						 databasePath,
-						 PGrnDatabaseBasename);
+	join_path_components(pgrnDatabasePath, databasePath, PGrnDatabaseBasename);
 
 	P(": flush: %u/%u", databaseOid, tableSpaceOid);
 
-	pgstat_report_activity(STATE_RUNNING, TAG ": flushing");
+	pgstat_report_activity(STATE_RUNNING, PGRN_TAG ": flushing");
 
-	grn_thread_set_get_limit_func(pgroonga_crash_safer_get_thread_limit,
-								  NULL);
-	grn_default_logger_set_flags(grn_default_logger_get_flags() |
-								 GRN_LOG_PID);
+	grn_thread_set_get_limit_func(pgroonga_crash_safer_get_thread_limit, NULL);
+	grn_default_logger_set_flags(grn_default_logger_get_flags() | GRN_LOG_PID);
 	grn_default_logger_set_max_level(PGroongaCrashSaferLogLevel);
 	if (!PGrnIsNoneValue(PGroongaCrashSaferLogPath))
 	{
@@ -403,7 +422,7 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		ereport(ERROR,
 				(errcode(ERRCODE_SYSTEM_ERROR),
-				 errmsg(TAG ": failed to initialize Groonga")));
+				 errmsg(PGRN_TAG ": failed to initialize Groonga")));
 	}
 
 	grn_set_segv_handler();
@@ -413,18 +432,18 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		ereport(ERROR,
 				(errcode(ERRCODE_SYSTEM_ERROR),
-				 errmsg(TAG ": failed to initialize Groonga context")));
+				 errmsg(PGRN_TAG ": failed to initialize Groonga context")));
 	}
 
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": initialize: <%s>: %u/%u",
+			PGRN_TAG ": initialize: <%s>: %u/%u",
 			PGRN_VERSION,
 			databaseOid,
 			tableSpaceOid);
 	GRN_LOG(&ctx,
 			GRN_LOG_DEBUG,
-			TAG ": max_recovery_threads: %d",
+			PGRN_TAG ": max_recovery_threads: %d",
 			grn_get_default_n_workers());
 
 	grn_ctx_set_wal_role(&ctx, GRN_WAL_ROLE_PRIMARY);
@@ -442,17 +461,18 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		GRN_LOG(&ctx,
 				GRN_LOG_WARNING,
-				TAG ": failed to %s database: <%s>",
+				PGRN_TAG ": failed to %s database: <%s>",
 				pgrnDatabasePathExist ? "open" : "create",
 				pgrnDatabasePath);
 		PGrnDatabaseRemoveAllRelatedFiles(databasePath);
 		db = grn_db_create(&ctx, pgrnDatabasePath, NULL);
 		if (!db)
 		{
-			ereport(ERROR,
-					(errcode(ERRCODE_SYSTEM_ERROR),
-					 errmsg(TAG ": failed to recreate Groonga database: %s",
-							ctx.errbuf)));
+			ereport(
+				ERROR,
+				(errcode(ERRCODE_SYSTEM_ERROR),
+				 errmsg(PGRN_TAG ": failed to recreate Groonga database: %s",
+						ctx.errbuf)));
 		}
 		needReindex = true;
 	}
@@ -465,29 +485,30 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 		GRN_LOG(&ctx,
 				GRN_LOG_NOTICE,
-				TAG ": reindexing: %u/%u",
+				PGRN_TAG ": %s: %u/%u",
+				needReindex ? "reindexing" : "resetting-position",
 				databaseOid,
 				tableSpaceOid);
 
 		snprintf(worker.bgw_name,
 				 BGW_MAXLEN,
-				 TAG ": prepare: %s: %u/%u",
+				 PGRN_TAG ": prepare: %s: %u/%u",
 				 needReindex ? "reindex" : "reset-position",
 				 databaseOid,
 				 tableSpaceOid);
 		snprintf(worker.bgw_type, BGW_MAXLEN, "%s", worker.bgw_name);
 		worker.bgw_flags =
-			BGWORKER_SHMEM_ACCESS |
-			BGWORKER_BACKEND_DATABASE_CONNECTION;
+			BGWORKER_SHMEM_ACCESS | BGWORKER_BACKEND_DATABASE_CONNECTION;
 		worker.bgw_start_time = BgWorkerStart_ConsistentState;
 		worker.bgw_restart_time = BGW_NEVER_RESTART;
 		snprintf(worker.bgw_library_name,
 				 BGW_MAXLEN,
-				 "%s", PGroongaCrashSaferLibraryName);
-		snprintf(worker.bgw_function_name, BGW_MAXLEN,
-				 needReindex ?
-				 "pgroonga_crash_safer_reindex_one" :
-				 "pgroonga_crash_safer_reset_position_one");
+				 "%s",
+				 PGroongaCrashSaferLibraryName);
+		snprintf(worker.bgw_function_name,
+				 BGW_MAXLEN,
+				 needReindex ? "pgroonga_crash_safer_reindex_one"
+							 : "pgroonga_crash_safer_reset_position_one");
 		worker.bgw_main_arg = databaseInfoDatum;
 		worker.bgw_notify_pid = MyProcPid;
 		if (RegisterDynamicBackgroundWorker(&worker, &handle))
@@ -495,7 +516,8 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 			WaitForBackgroundWorkerShutdown(handle);
 			GRN_LOG(&ctx,
 					GRN_LOG_NOTICE,
-					TAG ": reindexed: %u/%u",
+					PGRN_TAG ": %s: %u/%u",
+					needReindex ? "reindexed" : "reset-position",
 					databaseOid,
 					tableSpaceOid);
 		}
@@ -503,7 +525,7 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": ready: %u/%u",
+			PGRN_TAG ": ready: %u/%u",
 			databaseOid,
 			tableSpaceOid);
 
@@ -514,13 +536,10 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 
 	while (!PGroongaCrashSaferGotSIGTERM)
 	{
-		TimestampTz nextFlushTime =
-			TimestampTzPlusMilliseconds(
-				lastFlushTime,
-				PGroongaCrashSaferFlushNaptime * 1000);
-		long timeout =
-			PGrnTimestampDifferenceMilliseconds(GetCurrentTimestamp(),
-												nextFlushTime);
+		TimestampTz nextFlushTime = TimestampTzPlusMilliseconds(
+			lastFlushTime, PGroongaCrashSaferFlushNaptime * 1000);
+		long timeout = PGrnTimestampDifferenceMilliseconds(
+			GetCurrentTimestamp(), nextFlushTime);
 		int conditions;
 		if (timeout <= 0)
 		{
@@ -528,12 +547,11 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 		}
 		else
 		{
-			conditions = WaitLatch(MyLatch,
-								   WL_LATCH_SET |
-								   WL_TIMEOUT |
-								   PGRN_WL_EXIT_ON_PM_DEATH,
-								   timeout,
-								   PG_WAIT_EXTENSION);
+			conditions =
+				WaitLatch(MyLatch,
+						  WL_LATCH_SET | WL_TIMEOUT | PGRN_WL_EXIT_ON_PM_DEATH,
+						  timeout,
+						  PG_WAIT_EXTENSION);
 		}
 		if (conditions & WL_LATCH_SET)
 		{
@@ -578,24 +596,22 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		int conditions;
 		uint32 n_using_processes =
-			pgrn_crash_safer_statuses_get_n_using_processes(statuses,
-															databaseOid,
-															tableSpaceOid);
+			pgrn_crash_safer_statuses_get_n_using_processes(
+				statuses, databaseOid, tableSpaceOid);
 		if (n_using_processes == 0)
 			break;
 
 		GRN_LOG(&ctx,
 				GRN_LOG_NOTICE,
-				TAG ": waiting for connections to finish: %u: %u/%u",
+				PGRN_TAG ": waiting for connections to finish: %u: %u/%u",
 				n_using_processes,
 				databaseOid,
 				tableSpaceOid);
-		conditions = WaitLatch(MyLatch,
-							   WL_LATCH_SET |
-							   WL_TIMEOUT |
-							   PGRN_WL_EXIT_ON_PM_DEATH,
-							   PGroongaCrashSaferFlushNaptime * 1000,
-							   PG_WAIT_EXTENSION);
+		conditions =
+			WaitLatch(MyLatch,
+					  WL_LATCH_SET | WL_TIMEOUT | PGRN_WL_EXIT_ON_PM_DEATH,
+					  PGroongaCrashSaferFlushNaptime * 1000,
+					  PG_WAIT_EXTENSION);
 		if (conditions & WL_LATCH_SET)
 		{
 			ResetLatch(MyLatch);
@@ -613,26 +629,26 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	{
 		GRN_LOG(&ctx,
 				GRN_LOG_NOTICE,
-				TAG ": flushing database before closing: %u/%u",
+				PGRN_TAG ": flushing database before closing: %u/%u",
 				databaseOid,
 				tableSpaceOid);
 		grn_obj_flush_recursive(&ctx, db);
 		GRN_LOG(&ctx,
 				GRN_LOG_NOTICE,
-				TAG ": flushed database before closing: %u/%u",
+				PGRN_TAG ": flushed database before closing: %u/%u",
 				databaseOid,
 				tableSpaceOid);
 	}
 
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": closing database: %u/%u",
+			PGRN_TAG ": closing database: %u/%u",
 			databaseOid,
 			tableSpaceOid);
 	grn_obj_close(&ctx, db);
 	GRN_LOG(&ctx,
 			GRN_LOG_NOTICE,
-			TAG ": closed database: %u/%u",
+			PGRN_TAG ": closed database: %u/%u",
 			databaseOid,
 			tableSpaceOid);
 
@@ -655,15 +671,11 @@ pgroonga_crash_safer_main_flush_one(pgrn_crash_safer_statuses_entry *entry)
 	Oid databaseOid;
 	Oid tableSpaceOid;
 
-	PGRN_DATABASE_INFO_UNPACK(entry->key,
-							  databaseOid,
-							  tableSpaceOid);
-	P(": flush: start: %u/%u",
-	  databaseOid,
-	  tableSpaceOid);
+	PGRN_DATABASE_INFO_UNPACK(entry->key, databaseOid, tableSpaceOid);
+	P(": flush: start: %u/%u", databaseOid, tableSpaceOid);
 	snprintf(worker.bgw_name,
 			 BGW_MAXLEN,
-			 TAG ": flush: %u/%u",
+			 PGRN_TAG ": flush: %u/%u",
 			 databaseOid,
 			 tableSpaceOid);
 	snprintf(worker.bgw_type, BGW_MAXLEN, "%s", worker.bgw_name);
@@ -672,10 +684,10 @@ pgroonga_crash_safer_main_flush_one(pgrn_crash_safer_statuses_entry *entry)
 	worker.bgw_restart_time = BGW_NEVER_RESTART;
 	snprintf(worker.bgw_library_name,
 			 BGW_MAXLEN,
-			 "%s", PGroongaCrashSaferLibraryName);
-	snprintf(worker.bgw_function_name,
-			 BGW_MAXLEN,
-			 "pgroonga_crash_safer_flush_one");
+			 "%s",
+			 PGroongaCrashSaferLibraryName);
+	snprintf(
+		worker.bgw_function_name, BGW_MAXLEN, "pgroonga_crash_safer_flush_one");
 	worker.bgw_main_arg = DatumGetUInt64(entry->key);
 	worker.bgw_notify_pid = MyProcPid;
 	if (!RegisterDynamicBackgroundWorker(&worker, &handle))
@@ -695,7 +707,7 @@ pgroonga_crash_safer_main_flush_all(void)
 	StartTransactionCommand();
 	PushActiveSnapshot(GetTransactionSnapshot());
 	pgstat_report_activity(STATE_RUNNING,
-						   TAG ": start flush process for all databases");
+						   PGRN_TAG ": start flush process for all databases");
 
 	statuses = pgrn_crash_safer_statuses_get();
 
@@ -716,17 +728,13 @@ pgroonga_crash_safer_main_flush_all(void)
 		tableSpaceOid = form->dattablespace;
 
 		databasePath = GetDatabasePath(databaseOid, tableSpaceOid);
-		join_path_components(pgrnDatabasePath,
-							 databasePath,
-							 PGrnDatabaseBasename);
+		join_path_components(
+			pgrnDatabasePath, databasePath, PGrnDatabaseBasename);
 		if (!pgrn_file_exist(pgrnDatabasePath))
 			continue;
 
-		entry = pgrn_crash_safer_statuses_search(statuses,
-												 databaseOid,
-												 tableSpaceOid,
-												 HASH_ENTER,
-												 NULL);
+		entry = pgrn_crash_safer_statuses_search(
+			statuses, databaseOid, tableSpaceOid, HASH_ENTER, NULL);
 		pgroonga_crash_safer_main_flush_one(entry);
 	}
 	table_endscan(scan);
@@ -766,8 +774,7 @@ pgroonga_crash_safer_main(Datum arg)
 		int conditions;
 
 		conditions = WaitLatch(MyLatch,
-							   WL_LATCH_SET |
-							   PGRN_WL_EXIT_ON_PM_DEATH,
+							   WL_LATCH_SET | PGRN_WL_EXIT_ON_PM_DEATH,
 							   0,
 							   PG_WAIT_EXTENSION);
 		if (conditions & WL_LATCH_SET)
@@ -851,35 +858,42 @@ _PG_init(void)
 							 NULL,
 							 NULL);
 
-	DefineCustomIntVariable("pgroonga_crash_safer.max_recovery_threads",
-							"Maximum number of threads for recovery of broken Groonga indexes.",
-							"The default is 0, which means disabled. "
-							"Use all CPUs in the environment at -1. "
-							"Use CPU for that number if 1 or later is set.",
-							&PGroongaCrashSaferMaxRecoveryThreads,
-							PGroongaCrashSaferMaxRecoveryThreads,
-							-1,
-							INT_MAX,
-							PGC_SIGHUP,
-							0,
-							NULL,
-							NULL,
-							NULL);
+	DefineCustomIntVariable(
+		"pgroonga_crash_safer.max_recovery_threads",
+		"Maximum number of threads for recovery of broken Groonga indexes.",
+		"The default is 0, which means disabled. "
+		"Use all CPUs in the environment at -1. "
+		"Use CPU for that number if 1 or later is set.",
+		&PGroongaCrashSaferMaxRecoveryThreads,
+		PGroongaCrashSaferMaxRecoveryThreads,
+		-1,
+		INT_MAX,
+		PGC_SIGHUP,
+		0,
+		NULL,
+		NULL,
+		NULL);
 
 	if (!process_shared_preload_libraries_in_progress)
 		return;
 
-	snprintf(worker.bgw_name, BGW_MAXLEN, TAG ": main");
+#ifdef PGRN_SUPPORT_WAL_RESOURCE_MANAGER
+	/* Use pgroonga-wal-resource-manager for crash safe on standby. */
+	if (StandbyMode && RmgrIdExists(PGRN_WAL_RESOURCE_MANAGER_ID))
+		return;
+#endif
+
+	snprintf(worker.bgw_name, BGW_MAXLEN, PGRN_TAG ": main");
 	snprintf(worker.bgw_type, BGW_MAXLEN, "%s", worker.bgw_name);
 	worker.bgw_flags =
-		BGWORKER_SHMEM_ACCESS |
-		BGWORKER_BACKEND_DATABASE_CONNECTION;
+		BGWORKER_SHMEM_ACCESS | BGWORKER_BACKEND_DATABASE_CONNECTION;
 	worker.bgw_start_time = BgWorkerStart_ConsistentState;
 	worker.bgw_restart_time = 60;
-	snprintf(worker.bgw_library_name, BGW_MAXLEN,
-			 "%s", PGroongaCrashSaferLibraryName);
-	snprintf(worker.bgw_function_name, BGW_MAXLEN,
-			 "pgroonga_crash_safer_main");
+	snprintf(worker.bgw_library_name,
+			 BGW_MAXLEN,
+			 "%s",
+			 PGroongaCrashSaferLibraryName);
+	snprintf(worker.bgw_function_name, BGW_MAXLEN, "pgroonga_crash_safer_main");
 	worker.bgw_main_arg = 0;
 	worker.bgw_notify_pid = 0;
 
