@@ -811,10 +811,44 @@ pgrnwrm_redo(XLogReaderState *record)
 static void
 pgrnwrm_desc(StringInfo buffer, XLogReaderState *record)
 {
-	if (!StandbyMode)
-		return;
+	uint8 info;
 
-	GRN_LOG(ctx, GRN_LOG_DEBUG, PGRN_TAG ": [desc]");
+	if (StandbyMode)
+		GRN_LOG(ctx, GRN_LOG_DEBUG, PGRN_TAG ": [desc]");
+
+	appendStringInfo(buffer, "rmid: %d", XLogRecGetRmid(record));
+	/* TODO: Add more information */
+	info = XLogRecGetInfo(record) & XLR_RMGR_INFO_MASK;
+	switch (info)
+	{
+	case PGRN_WAL_RECORD_CREATE_TABLE:
+		appendStringInfo(buffer, " action: create-table");
+		break;
+	case PGRN_WAL_RECORD_CREATE_COLUMN:
+		appendStringInfo(buffer, " action: create-column");
+		break;
+	case PGRN_WAL_RECORD_SET_SOURCES:
+		appendStringInfo(buffer, " action: set-sources");
+		break;
+	case PGRN_WAL_RECORD_RENAME_TABLE:
+		appendStringInfo(buffer, " action: rename-table");
+		break;
+	case PGRN_WAL_RECORD_INSERT:
+		appendStringInfo(buffer, " action: insert");
+		break;
+	case PGRN_WAL_RECORD_DELETE:
+		appendStringInfo(buffer, " action: delete");
+		break;
+	case PGRN_WAL_RECORD_REMOVE_OBJECT:
+		appendStringInfo(buffer, " action: remove-object");
+		break;
+	case PGRN_WAL_RECORD_REGISTER_PLUGIN:
+		appendStringInfo(buffer, " action: register-plugin");
+		break;
+	default:
+		appendStringInfo(buffer, " action: unknown(%u)", info);
+		break;
+	}
 }
 
 static const char *
@@ -884,25 +918,6 @@ pgrnwrm_cleanup(void)
 	grn_fin();
 }
 
-static void
-pgrnwrm_mask(char *pagedata, BlockNumber block_number)
-{
-	if (!StandbyMode)
-		return;
-
-	GRN_LOG(ctx, GRN_LOG_DEBUG, PGRN_TAG ": mask");
-}
-
-static void
-pgrnwrm_decode(struct LogicalDecodingContext *context,
-			   struct XLogRecordBuffer *buffer)
-{
-	if (!StandbyMode)
-		return;
-
-	GRN_LOG(ctx, GRN_LOG_DEBUG, PGRN_TAG ": decode");
-}
-
 static RmgrData data = {
 	.rm_name = "PGroonga",
 	.rm_redo = pgrnwrm_redo,
@@ -910,8 +925,8 @@ static RmgrData data = {
 	.rm_identify = pgrnwrm_identify,
 	.rm_startup = pgrnwrm_startup,
 	.rm_cleanup = pgrnwrm_cleanup,
-	.rm_mask = pgrnwrm_mask,
-	.rm_decode = pgrnwrm_decode,
+	.rm_mask = NULL,
+	.rm_decode = NULL,
 };
 #endif
 
